#!/usr/bin/env python3

from math import acos, atan2, cos, degrees, radians, sin, sqrt
from typing import Tuple
import struct

OBLIQUITY = radians(23.4392911)
SIN_OB = sin(OBLIQUITY)
COS_OB = cos(OBLIQUITY)

SP_TYPES = [
    'O', 'B', 'A', 'F', 'G', 'K', 'M', 'R',
    'S', 'N', 'WC', 'WN', '?', 'L', 'T', 'C',
]

WD_TYPES = [
    'DA', 'DB', 'DC', 'DO', 'DQ', 'DZ', 'D', 'DX',
]

LUM_TYPES = [
    'Ia0', 'Ia', 'Ib', 'II', 'III', 'IV', 'V', 'VI',
]

def spec_str(code: int) -> str:
    """Convert a spectral type code to string."""
    k = (code & 0xf000) >> 12
    if k == 0:
        t = (code & 0x0f00) >> 8
        result = SP_TYPES[t]
        if t != 0xc:
            s = (code & 0x00f0) >> 4
            if s < 10:
                result += str(s)
            lum = code & 0x000f
            if lum < len(LUM_TYPES):
                result += LUM_TYPES[lum]
    elif k == 1:
        t = (code & 0x0f00) >> 8
        if t < len(WD_TYPES):
            result = WD_TYPES[t]
            s = (code & 0x00f0) >> 4
            if s < 10:
                result += str(s)
        else:
            result = '?'
    elif k == 2:
        result = 'Q'
    elif k == 3:
        result = 'X'
    else:
        result = '?'
    return result

def xyz_to_ang(x: float, y: float, z: float) -> Tuple[float, float, float]:
    """Convert ecliptic Cartesian coordinates to RA/Dec/Dist"""
    yy = COS_OB*y - SIN_OB*z
    zz = SIN_OB*y + COS_OB*z
    dist = sqrt(x*x + yy*yy + zz*zz)
    ra = degrees(atan2(-zz, x))
    if ra < 0:
        ra += 360
    dec = 90-degrees(acos(yy/dist))
    return ra, dec, dist

with open('stars.dat', 'rb') as f:
    header = f.read(14)
    ftype, fversion, nrecords = struct.unpack('<8sHL', header)
    if ftype != b'CELSTARS' or fversion != 0x0100:
        raise RuntimeError('Unknown file format')
    print('HIP,"x (ly)","y (ly)","z (ly)","RA (deg)","Dec (deg)","Dist (ly)",V_abs,spec_type')
    fmt = struct.Struct('<L3fhH')
    for i in range(nrecords):
        record = f.read(20)
        hip, x, y, z, vmag, spec = fmt.unpack(record)
        vmag /= 256
        ra, dec, dist = xyz_to_ang(x, y, z)
        print(f'{hip},{x},{y},{z},{ra},{dec},{dist},{vmag},{spec_str(spec)}')
